from jupyter_ascending.requests.execute import _find_cell_number


def test_find_cell_number():
    assert (
        _find_cell_number(
            [
                "\n",
                "# %%\n",
                "print('hello')\n",
                "print('wow!')",
                "# %%\n" "TARGET = True",
                "\n",
                "# %%",
                "# Too Far",
                "# Still too far",
            ],
            5,
        )
        == 1
    )


def test_find_cell_number_with_more_spaces():
    assert (
        _find_cell_number(
            [
                "\n",
                "#     %%\n",
                "print('hello')\n",
                "print('wow!')",
                "# %%\n" "TARGET = True",
                "\n",
                "# %%",
                "# Too Far",
                "# Still too far",
            ],
            5,
        )
        == 1
    )



from pathlib import Path
from typing import Any
from typing import Dict
from typing import Optional

from jsonrpcclient.clients.tornado_client import TornadoClient
from jsonrpcserver import async_dispatch as dispatch
from jsonrpcserver import method
from loguru import logger
from notebook.base.handlers import IPythonHandler  # type: ignore
from notebook.utils import url_path_join  # type: ignore

from jupyter_ascending._environment import SYNC_EXTENSION
from jupyter_ascending.errors import UnableToFindNotebookException
from jupyter_ascending.functional import get_matching_tail_tokens

_REGISTERED_SERVERS: Dict[str, int] = {}


def _clear_registered_servers():
    global _REGISTERED_SERVERS

    _REGISTERED_SERVERS = {}


class JupyterAscendingHandler(IPythonHandler):
    async def post(self) -> None:
        """We receive commands as HTTP POST requests.

        NOTE: authentication is disabled on this endpoint!!!

        It is critical that this doesn't block the main server thread,
        or you'll get a deadlock in the notebook kernel thread as it processes
        this request. Thus the usage of `asyncio`.
        """
        request = self.request.body.decode()
        response = await dispatch(request)
        logger.info("Got Response:\n\t\t{}", response)
        self.write(str(response))

    def check_xsrf_cookie(self):
        """Disable XSRF cookie checking on this request type"""


def load_extension(nb_server_app):
    """
    Called when the extension is loaded.

    Args:
        nb_server_app (NotebookWebApplication): handle to the Notebook webserver instance.
    """
    web_app = nb_server_app.web_app
    host_pattern = ".*$"
    route_pattern = url_path_join(web_app.settings["base_url"], "/jupyter_ascending")
    web_app.add_handlers(host_pattern, [(route_pattern, JupyterAscendingHandler)])


@method
async def register_notebook_server(notebook_path: str, port_number: int) -> None:
    logger.info("Registering notebook {notebook} on port {port}", notebook=notebook_path, port=port_number)

    _REGISTERED_SERVERS[notebook_path] = port_number

    logger.debug("Updated notebook mappings: {}", _REGISTERED_SERVERS)


@method
async def perform_notebook_request(notebook_path: str, command_name: str, data: Dict[str, Any]) -> Optional[Dict]:
    """Receives a command from the client library, picks the notebook that matches
    the filepath, and forwards the command along to that notebook."""
    logger.debug("Performing notebook request... ")

    try:
        notebook_server = get_server_for_notebook(notebook_path)
    except UnableToFindNotebookException:
        message = f"""Unable to find a paired notebook for {notebook_path} in registered notebooks: {_REGISTERED_SERVERS}.
Either a properly named notebook (ending in .sync.ipynb) is not running, or it didn't register properly for some reason."""
        logger.warning(message)
        return {"success": False, "error": message}

    client = TornadoClient(notebook_server)
    response = await client.request(command_name, data=data)
    if not response.data.ok:
        message = "Got failed response from notebook: {response}"
        logger.error(message)
        return {"success": False, "error": message}
    return {"success": True}


def _make_url(notebook_port: int):
    return f"http://localhost:{notebook_port}"


def get_server_for_notebook(notebook_str: str) -> Optional[str]:
    """Get the URL to the server running on the Jupyter notebook that best matches this filename."""
    # Normalize to notebook path
    notebook_str = notebook_str.replace(f".{SYNC_EXTENSION}.py", f".{SYNC_EXTENSION}.ipynb")
    logger.debug("Finding server for notebook_str, script_path: {}", notebook_str)

    notebook_path = Path(notebook_str)

    def get_score_for_name(registered_name: str) -> int:
        """
        Note that it is matching on parts of a path

        Returns the consecutive count of matching parts of a path, from the end toward the start.

        registered ['tmp', 'notebooks', 'myfile.py']
        notebook   ['opt', 'notebooks', 'myfile.py']
         -> 2

        registered ['a', 'b', 'c']
        notebook   ['a', 'b', 'd']
         -> 0

        """
        return len(get_matching_tail_tokens(notebook_path.parts, Path(registered_name).parts))

    score_by_name = {x: get_score_for_name(x) for x in _REGISTERED_SERVERS.keys()}

    if len(score_by_name) == 0:
        raise UnableToFindNotebookException(f"No registered notebooks")

    max_score = max(score_by_name.values())

    if max_score <= 0:
        raise UnableToFindNotebookException(f"Could not find server for notebook_str: {notebook_str}")

    # Only found one reasonable notebook.
    best_scores = [k for k, v in score_by_name.items() if v == max_score]

    if len(best_scores) == 1:
        notebook_port = _REGISTERED_SERVERS[best_scores[0]]

        logger.debug("Found server at port {}", notebook_port)
        return _make_url(notebook_port)
    else:
        raise UnableToFindNotebookException(f"Could not find server for notebook_str: {notebook_str}")



import pytest

from jupyter_ascending._environment import SYNC_EXTENSION
from jupyter_ascending.errors import UnableToFindNotebookException
from jupyter_ascending.handlers.server_extension import _clear_registered_servers
from jupyter_ascending.handlers.server_extension import _make_url
from jupyter_ascending.handlers.server_extension import get_server_for_notebook
from jupyter_ascending.handlers.server_extension import register_notebook_server


class TestGetSever:
    def setup_method(self, _method):
        _clear_registered_servers()

    @pytest.mark.asyncio
    async def test_clear_registered_servers(self):
        notebook_name = f"hello.{SYNC_EXTENSION}.ipynb"
        await register_notebook_server(notebook_name, 1234)
        assert get_server_for_notebook(notebook_name) is not None

        _clear_registered_servers()
        with pytest.raises(UnableToFindNotebookException):
            get_server_for_notebook(notebook_name)

    @pytest.mark.asyncio
    async def test_exact_match(self):
        notebook_name = f"hello.{SYNC_EXTENSION}.ipynb"
        notebook_port = 1234

        await register_notebook_server(notebook_name, notebook_port)
        assert get_server_for_notebook(notebook_name) == _make_url(notebook_port)

    @pytest.mark.asyncio
    async def test_stem_match(self):
        true_notebook_name = f"/home/tj/git/notebook.{SYNC_EXTENSION}.ipynb"
        remote_notebook_name = f"/home/other/git/notebook.{SYNC_EXTENSION}.ipynb"

        notebook_port = 1234

        await register_notebook_server(true_notebook_name, notebook_port)
        assert get_server_for_notebook(true_notebook_name) == _make_url(notebook_port)
        assert get_server_for_notebook(remote_notebook_name) == _make_url(notebook_port)

    @pytest.mark.asyncio
    async def test_more_than_stem_match(self):
        true_notebook_name = f"/home/tj/git/notebook.{SYNC_EXTENSION}.ipynb"
        do_not_pick_notebook = f"/home/tj/do_not_pick/notebook.{SYNC_EXTENSION}.ipynb"
        remote_notebook_name = f"/home/other/git/notebook.{SYNC_EXTENSION}.ipynb"

        notebook_port = 1234
        do_not_pick_port = 4444

        await register_notebook_server(true_notebook_name, notebook_port)
        await register_notebook_server(do_not_pick_notebook, do_not_pick_port)

        assert get_server_for_notebook(true_notebook_name) == _make_url(notebook_port)
        assert get_server_for_notebook(remote_notebook_name) == _make_url(notebook_port)

    @pytest.mark.asyncio
    async def test_equally_matching_stem_errors(self):
        foo_notebook_name = f"/home/foo/git/notebook.{SYNC_EXTENSION}.ipynb"
        bar_notebook_name = f"/home/bar/git/notebook.{SYNC_EXTENSION}.ipynb"

        foo_port = 1234
        bar_port = 4321

        current_notebook_name = f"/home/tj/git/notebook.{SYNC_EXTENSION}.ipynb"

        await register_notebook_server(foo_notebook_name, foo_port)
        await register_notebook_server(bar_notebook_name, bar_port)

        assert get_server_for_notebook(foo_notebook_name) == _make_url(foo_port)
        assert get_server_for_notebook(bar_notebook_name) == _make_url(bar_port)

        # Can't determine what this server
        with pytest.raises(UnableToFindNotebookException):
            get_server_for_notebook(current_notebook_name)

    @pytest.mark.asyncio
    async def test_no_matches(self):
        loaded_notebook_name = f"/home/tj/notebook.{SYNC_EXTENSION}.ipynb"
        not_loaded_notebook_name = f"/home/tj/other.{SYNC_EXTENSION}.ipynb"

        await register_notebook_server(loaded_notebook_name, 1234)

        with pytest.raises(UnableToFindNotebookException):
            get_server_for_notebook(not_loaded_notebook_name)



"""Defines messages and data payloads for the JSON-RPC client-server protocol
+used to send our custom messages to the Jupyter notebook.

TODO: dataclass looks to be depreciated. Replace with attr.define"""
from typing import Optional

from attr import dataclass


@dataclass
class JsonBaseRequest:
    file_name: str


@dataclass
class ExecuteRequest(JsonBaseRequest):
    cell_index: int
    contents: Optional[str]


@dataclass
class ExecuteAllRequest(JsonBaseRequest):
    pass


@dataclass
class RestartRequest(JsonBaseRequest):
    pass


@dataclass
class SyncRequest(JsonBaseRequest):
    contents: str


@dataclass
class GetStatusRequest(JsonBaseRequest):
    pass


@dataclass
class FocusCellRequest(JsonBaseRequest):
    cell_index: int



from itertools import takewhile
from typing import Callable
from typing import List
from typing import Sequence


def get_matching_tail_tokens(a: Sequence, b: Sequence) -> List:
    return list(takewhile(lambda parts: parts[0] == parts[1], zip(reversed(a), reversed(b))))


# TODO: This was just as an example, idk if it works.
def compose(*args: Callable) -> Callable:
    def _func(result):
        for f in reversed(args):
            result = f(result)

        return result

    return _func



from typing import List

import attr

from jupyter_ascending.notebook.data_types import JupyterCell
from jupyter_ascending.notebook.data_types import NotebookContents


def evolve_notebook_cells(contents: NotebookContents, new_cells: List[JupyterCell]) -> NotebookContents:
    assert len(new_cells) == len(set(x.index for x in new_cells)), "Must have unique indeces"

    return attr.evolve(contents, cells=list(sorted(new_cells, key=lambda x: x.index)))


def evolve_cell_source(contents: NotebookContents, index: int, source: List[str]) -> NotebookContents:
    new_cells: List[JupyterCell] = []
    for cell in contents.cells:
        if cell.index == index:
            new_cells.append(attr.evolve(cell, source=source))
        else:
            new_cells.append(cell)

    return evolve_notebook_cells(contents, new_cells)


def evolve_cell_type(contents: NotebookContents, index: int, cell_type: str) -> NotebookContents:
    new_cells: List[JupyterCell] = []
    for cell in contents.cells:
        if cell.index == index:
            new_cells.append(attr.evolve(cell, cell_type=cell_type))
        else:
            new_cells.append(cell)

    return evolve_notebook_cells(contents, new_cells)



import argparse
from pathlib import Path

from loguru import logger

from jupyter_ascending._environment import SYNC_EXTENSION
from jupyter_ascending.json_requests import SyncRequest
from jupyter_ascending.logger import setup_logger
from jupyter_ascending.requests.client_lib import request_notebook_command


@logger.catch
def send(file_name: str):
    if f".{SYNC_EXTENSION}.py" not in file_name:
        return

    logger.info(f"Syncing File: {file_name}...")
    file_name = str(Path(file_name).absolute())

    with open(file_name, "r") as reader:
        raw_result = reader.read()

    request_obj = SyncRequest(file_name=file_name, contents=raw_result)
    request_notebook_command(request_obj)

    logger.info("... Complete")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    setup_logger()

    parser.add_argument("--filename", help="Filename to send")

    arguments = parser.parse_args()
    send(arguments.filename)



import os

EXECUTE_HOST = os.getenv("JUPYTER_ASCENDING_EXECUTE_HOST", "localhost")
EXECUTE_PORT = os.getenv("JUPYTER_ASCENDING_EXECUTE_PORT", 8888)

EXECUTE_HOST_LOCATION = (EXECUTE_HOST, EXECUTE_PORT)
EXECUTE_HOST_URL = f"http://{EXECUTE_HOST_LOCATION[0]}:{EXECUTE_HOST_LOCATION[1]}/jupyter_ascending"

LOG_LEVEL = os.getenv("JUPYTER_ASCENDING_LOG_LEVEL", "INFO")
SHOW_TO_STDOUT = os.getenv("JUPYTER_ASCENDING_SHOW_TO_STDOUT", False)
# TODO: it would be great for this to be an environment variable... but unfortunately we need to know the value
#  on the javascript side as well, and I'm not sure how to get this value over there easily. Would love help!
SYNC_EXTENSION = "sync"



import argparse
import re
from functools import partial
from pathlib import Path
from typing import List

from loguru import logger

from jupyter_ascending.json_requests import ExecuteRequest
from jupyter_ascending.logger import setup_logger
from jupyter_ascending.requests.client_lib import request_notebook_command
from jupyter_ascending.requests.sync import send as sync_send

CELL_SEPARATOR_PATTERNS = [
    re.compile(r"#\s*%%"),
    re.compile(r"#\s*\+\+"),
]


def _find_cell_number(lines: List[str], line_number: int) -> int:
    # We need to split cells the same way that jupytext does so that our cell numbers line up.
    # Unfortunately there's not an obvious way to just use the jupytext parser.

    # The default case has a # %% on the first line. The first cell starts after this.
    # A second case has no # %% before code begins. The first cell starts immediately.
    # A third case has a single blank line before the # %%. The blank line is its own cell.
    if any(pat.match(lines[0]) for pat in CELL_SEPARATOR_PATTERNS):
        cell_index = -1
    else:
        cell_index = 0

    for index, line in enumerate(lines):
        if any(pat.match(line) for pat in CELL_SEPARATOR_PATTERNS):
            logger.debug(f"Found another new cell on line number: {index}")
            cell_index += 1
            logger.debug(f"    New cell index {cell_index}")

        # Found line number, quit
        if index == int(line_number):
            break

    return cell_index


def send(file_name: str, line_number: int, *args, **kwargs):
    logger.debug("Starting execute request")

    # Always pass absolute path
    file_name = str(Path(file_name).absolute())

    request_obj = partial(ExecuteRequest, file_name=file_name, contents="")

    with open(file_name, "r") as reader:
        lines = reader.readlines()

    cell_index = _find_cell_number(lines, line_number)

    final_request = request_obj(cell_index=cell_index)
    logger.info(f"Sending request with {final_request}")
    request_notebook_command(final_request)
    logger.info("... Complete")


if __name__ == "__main__":
    setup_logger()
    parser = argparse.ArgumentParser()

    parser.add_argument("--filename", help="Filename to send")
    parser.add_argument("--linenumber", help="Line number that the cursor is currently on")

    arguments = parser.parse_args()

    # Sync code first
    sync_send(arguments.filename)
    send(arguments.filename, arguments.linenumber)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) tjdevries.
# Distributed under the terms of the Modified BSD License.

from jupyter_ascending._version import __version__
from jupyter_ascending._version import version_info
from jupyter_ascending.extension import load_ipython_extension
from jupyter_ascending.extension import load_jupyter_server_extension
from jupyter_ascending.nbextension import _jupyter_nbextension_paths


def _jupyter_server_extension_paths():
    return [{
        "module": "jupyter_ascending",
    }]


__all__ = [
    "__version__",
    "_jupyter_nbextension_paths",
    "load_ipython_extension",
    "_jupyter_server_extension_paths",
    "load_jupyter_server_extension",
    "version_info",
]



"""
Make a pair of empty .sync.py and .sync.ipynb files.
"""
import argparse
from pathlib import Path

import jupytext

from jupyter_ascending._environment import SYNC_EXTENSION

_STARTER_CONTENTS = """# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.3.4
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %%
"""


def create_new_file(base: str, force: bool):
    assert not base.endswith(".py"), "base: Cannot end with '.py'"
    assert not base.endswith(
        f".{SYNC_EXTENSION}.py"
    ), f"base: Cannot end with '{SYNC_EXTENSION}.py' -- we're adding that!"
    assert not base.endswith(".ipynb"), "base: Cannot end with '.ipynb'"
    assert not base.endswith(f".{SYNC_EXTENSION}.ipynb"), "base: Cannot end with '.ipynb' -- we're adding that!"
    assert not base.endswith(f".{SYNC_EXTENSION}"), f"we're going to add '.{SYNC_EXTENSION}'"

    py_path = base + f".{SYNC_EXTENSION}.py"
    ipynb_path = base + f".{SYNC_EXTENSION}.ipynb"

    if not force and Path(py_path).exists():
        print(f"Path '{py_path}' already exists. Call with --force to override.")
        return

    if not force and Path(ipynb_path).exists():
        print(f"Path '{ipynb_path}' already exists. Call with --force to override.")
        return

    with open(py_path, "w") as f:
        print("Writing :", py_path)
        f.write(_STARTER_CONTENTS)

    print("Writing :", ipynb_path)
    jupytext.write(jupytext.reads(_STARTER_CONTENTS, "py:percent"), ipynb_path)
    # with open(ipynb_path, "w") as f:

    print("Success!")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--base", help="Base filename to add (do not include .py or .ipynb)", required=True)
    parser.add_argument("-f", "--force", help="Override existing files if passed.", default=False, action="store_true")

    arguments = parser.parse_args()
    create_new_file(arguments.base, arguments.force)



"""
This file contains code for the JSON-RPC server we run alongside each .sync.ipynb notebook.

It receives messages from `jupyter_server.py` and takes the appropriate action in the notebook.
"""
import threading
import time
from http.server import HTTPServer
from inspect import signature
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

import ipywidgets as widgets
import jupytext
from IPython.display import display
from ipykernel.comm import Comm
from jsonrpcclient import request
from loguru import logger

from jupyter_ascending._environment import EXECUTE_HOST_URL
from jupyter_ascending.handlers import ServerMethods
from jupyter_ascending.handlers import generate_request_handler
from jupyter_ascending.handlers.server_extension import register_notebook_server
from jupyter_ascending.json_requests import ExecuteAllRequest
from jupyter_ascending.json_requests import ExecuteRequest
from jupyter_ascending.json_requests import FocusCellRequest
from jupyter_ascending.json_requests import GetStatusRequest
from jupyter_ascending.json_requests import RestartRequest
from jupyter_ascending.json_requests import SyncRequest
from jupyter_ascending.notebook.data_types import JupyterCell
from jupyter_ascending.notebook.data_types import NotebookContents
from jupyter_ascending.notebook.merge import OpCodeAction
from jupyter_ascending.notebook.merge import OpCodes
from jupyter_ascending.notebook.merge import opcode_merge_cell_contents
from jupyter_ascending.utils import find_free_port

COMM_NAME = "AUTO_SYNC::notebook"

notebook_server_methods = ServerMethods("JupyterNotebook Start", "JupyterNotebook Close")

merge_complete = False

lock = threading.Lock()


@logger.catch
def start_notebook_server_in_thread(notebook_name: str, status_widget=None):
    """
    Args:
        notebook_name: The name of the notebook you want to be syncing in this process.
    """

    notebook_path = Path(notebook_name).absolute()

    if not status_widget:
        status_widget = widgets.Text()
        status_widget.style.description_width = "300px"
        display(status_widget)

    # TODO: This might be a race condition if a bunch of these started at once...
    notebook_server_port = find_free_port()

    notebook_executor = HTTPServer(("localhost", notebook_server_port), NotebookKernelRequestHandler,)
    notebook_executor_thread = threading.Thread(target=notebook_executor.serve_forever, args=tuple())
    notebook_executor_thread.start()

    logger.info("IPYTHON: Registering notebook {}", notebook_path)
    request(
        EXECUTE_HOST_URL,
        register_notebook_server.__name__,
        # Params
        notebook_path=str(notebook_path),
        port_number=notebook_server_port,
    )
    logger.info("==> Success")

    return status_widget


def dispatch_json_request(f):
    """
    A kinda weird decorator attempting to remove some boilerplate in the following funcs.
    Automatically dispatch a json request based on the request_type.

    Adds it to notebook_server_methods
    """

    # Get the type from the first argument of the function.
    #   This will define the name that we use to generate the method handling.
    request_type = signature(f).parameters["request_type"].annotation.__args__[0]

    def wrapped(data: Dict) -> str:
        return f(request_type, data)

    wrapped.__name__ = request_type.__name__

    return notebook_server_methods.add(wrapped)


@dispatch_json_request
def handle_execute_request(request_type: Type[ExecuteRequest], data: dict) -> str:
    """JSON-RPC request handler for 'execute cell'"""
    request = request_type(**data)

    comm = make_comm()
    execute_cell_contents(comm, request.cell_index)

    return f"Executing cell `{request.cell_index}`"


@dispatch_json_request
def handle_execute_all_request(request_type: Type[ExecuteAllRequest], data: dict) -> str:
    """JSON-RPC request handler for 'execute all cells'"""
    request = request_type(**data)

    # TODO: Remind myself why I don't need to say the filename here...
    comm = make_comm()
    execute_all_cells(comm)

    return f"Executing all cells in {request.file_name}"


@dispatch_json_request
def handle_sync_request(request_type: Type[SyncRequest], data: dict) -> str:
    """JSON-RPC request handler for 'sync'"""
    global merge_complete
    merge_complete = False
    request = request_type(**data)

    # We lock here because updating the notebook isn't threadsafe.
    # If we got two sync requests simultaneously without a lock,
    # bad things might happen (eg duplicated inserts/deletes).
    with lock:
        comm = make_comm()

        result = jupytext.reads(request.contents, fmt="py:percent")
        update_cell_contents(comm, result)

    return "Syncing all cells"


@dispatch_json_request
def handle_focus_cell_request(request_type: Type[FocusCellRequest], data: dict) -> str:
    """JSON-RPC request handler for 'focus cell'"""
    request = request_type(**data)

    print(request)
    raise NotImplementedError


@dispatch_json_request
def handle_get_status_request(request_type: Type[GetStatusRequest], data: dict) -> str:
    """JSON-RPC request handler for 'get status'"""
    logger.info("Attempting get_status")

    comm = make_comm()
    comm.send({"command": "get_status"})

    logger.info("Sent get_status")

    return f"Updating status"


@dispatch_json_request
def handle_restart_request(request_type: Type[RestartRequest], data: dict) -> str:
    """JSON-RPC request handler for 'restart'"""
    request = request_type(**data)

    comm = make_comm()
    comm.send({"command": "restart_kernel"})
    logger.info("Sent restart_kernel")

    return f"Restarting kernel in {request.file_name}"


NotebookKernelRequestHandler = generate_request_handler("NotebookKernel", notebook_server_methods)


def make_comm():
    """A comm is a Jupyter object for communicating between a notebook and kernel.

    Set up this object with event handlers."""
    global merge_complete

    logger.info("IPYTHON: Registering Comms")

    comm_target_name = COMM_NAME

    jupyter_comm = Comm(target_name=comm_target_name)

    def _get_command(msg) -> Optional[str]:
        return msg["content"]["data"].get("command", None)

    @jupyter_comm.on_msg
    def _recv(msg):
        global merge_complete
        if _get_command(msg) == "merge_notebooks":
            logger.info("GOT UPDATE STATUS")
            merge_notebooks(jupyter_comm, msg["content"]["data"])
            return

        if _get_command(msg) == "merge_complete":
            logger.info("GOT MERGE COMPLETE")
            merge_complete = True
            return

        logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
        logger.info(msg)
        logger.info("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")

    logger.info("==> Success")

    return jupyter_comm


def update_cell_contents(comm: Comm, result: Dict[str, Any]) -> None:
    # logger.info(Javascript("Jupyter.notebook.get_cells()"))
    def _transform_jupytext_cells(jupytext_cells) -> List[Dict[str, Any]]:
        """TODO: what does this do?"""
        return [
            {"index": i, "output": [], **{k: v for (k, v) in x.items() if k not in {"outputs", "metadata"}}}
            for i, x in enumerate(result["cells"])
        ]

    comm.send({"command": "start_sync_notebook", "cells": _transform_jupytext_cells(result["cells"])})

    # Wait for the merge_complete flag to get set in the callback.
    # This way we don't release the lock before syncing is done.
    for i in range(500):
        if merge_complete:
            return
        time.sleep(0.01)
    else:
        logger.warning("Timed out waiting for syncing to complete.")
    # contents = NotebookContents(cells=result["cells"])


def get_output_text(javascript_cell) -> Optional[str]:
    """Get cell output or return None if no output?"""
    output_tuple = javascript_cell.get("outputs", tuple())
    if not output_tuple:
        return None

    output = output_tuple[0]

    if output.get("data", None):
        data = output["data"]

        if isinstance(data, dict):
            if data.get("text/plain", None):
                return data["text/plain"]

    if output.get("text", None):
        return output["text"]

    return None


@logger.catch(reraise=True)
def merge_notebooks(comm: Comm, result: Dict[str, Any]) -> None:
    javascript_cells = result["javascript_cells"]
    current_notebook = NotebookContents(
        cells=[
            JupyterCell(
                index=i,
                cell_type=x["cell_type"],
                source=x["source"],
                output=get_output_text(x),
                # metadata=x["metadata"],
            )
            for i, x in enumerate(javascript_cells)
        ]
    )

    new_notebook = NotebookContents(cells=[JupyterCell(**x) for x in result["new_notebook"]])

    opcodes = opcode_merge_cell_contents(current_notebook, new_notebook)
    logger.info("Performing Opcodes...")
    logger.info(opcodes)

    net_shift = 0
    for op_action in opcodes:
        net_shift = perform_op_code(comm, op_action, current_notebook, new_notebook, net_shift)

    logger.info("sending finish_merge command")
    comm.send({"command": "finish_merge"})


def perform_op_code(
    comm: Comm,
    op_action: OpCodeAction,
    current_notebook: NotebookContents,
    updated_notebook: NotebookContents,
    net_shift: int,
) -> int:
    """
    net_shift (int): Tracks the net shift of previous op codes since we can't apply all the operations at the same time to jupyter,
                        since it does not have that kind of editting model.

                        So what we do is make sure that as we delete and insert, we keep track of the shifts that have happened thus far.
                        Given this shift, we will shift the actions that we tell Jupyter notebook to do.
    """

    if op_action.op_code == OpCodes.EQUAL:
        pass

    elif op_action.op_code == OpCodes.DELETE:
        logger.info(f"Performing Delete: {op_action}")

        # Since deletion is a bit goofy for jupyter, so it has to be adjusted by net shift thus far.
        cells_to_delete = [x + net_shift for x in range(*op_action.current)]
        comm.send({"command": "op_code__delete_cells", "cell_indices": cells_to_delete})

        net_shift = net_shift - len(cells_to_delete)

    elif op_action.op_code == OpCodes.INSERT:
        logger.info(f"Performing Insert: {op_action}")

        cells_to_insert = list(range(*op_action.updated))
        for cell_number in cells_to_insert:
            comm.send(
                {
                    "command": "op_code__insert_cell",
                    "cell_number": cell_number,
                    "cell_type": updated_notebook.cells[cell_number].cell_type,
                    "cell_contents": updated_notebook.cells[cell_number].joined_source,
                }
            )

        net_shift = net_shift + len(cells_to_insert)

    elif op_action.op_code == OpCodes.REPLACE:
        # Keep track of what the current cells looked like before.
        current_cells = list(range(*op_action.current))
        updated_cells = list(range(*op_action.updated))

        for cell_number in updated_cells:
            # If we have current cells we're replacing, do that.
            if current_cells:
                current_cells.pop(0)

                comm.send(
                    {
                        "command": "op_code__replace_cell",
                        "cell_number": cell_number,
                        "cell_type": updated_notebook.cells[cell_number].cell_type,
                        "cell_contents": updated_notebook.cells[cell_number].joined_source,
                    }
                )
            # Otherwise, we have new cells to insert so we don't overwrite existing cells
            else:
                net_shift = perform_op_code(
                    comm,
                    OpCodeAction(
                        op_code=OpCodes.INSERT,
                        # NOTE: This is intentionally the last index for both of these
                        current_start_idx=op_action.current_final_idx,
                        current_final_idx=op_action.current_final_idx,
                        updated_start_idx=cell_number,
                        updated_final_idx=cell_number + 1,
                    ),
                    current_notebook,
                    updated_notebook,
                    net_shift,
                )

        # If we have cells left over from the replace (i.e. 1-4 replaced with 1-2),
        #   then we need to delete the rest of them.
        if current_cells:
            net_shift = perform_op_code(
                comm,
                OpCodeAction(
                    op_code=OpCodes.DELETE,
                    current_start_idx=current_cells[0],
                    current_final_idx=current_cells[-1] + 1,
                    # NOTE: This is intentionally the last index for both of these
                    updated_start_idx=op_action.updated_final_idx,
                    updated_final_idx=op_action.updated_final_idx,
                ),
                current_notebook,
                updated_notebook,
                net_shift,
            )

    else:
        raise NotImplementedError

    return net_shift


def execute_cell_contents(comm: Comm, cell_number: int) -> None:
    comm.send({"command": "execute", "cell_number": cell_number})


def execute_all_cells(comm: Comm) -> None:
    comm.send({"command": "execute_all"})



from typing import Any
from typing import List
from typing import Optional
from typing import Tuple

import attr


def optional_tuple(val) -> Optional[Tuple[Any, ...]]:
    if val is None:
        return None

    if isinstance(val, str):
        return (val,)

    return tuple(val)


# def frozen_dict(val) -> Optional[


@attr.s(frozen=True)
class JupyterCell:
    # TODO: Should probably be an enum
    cell_type: str = attr.ib()

    #: -1 for unknown
    #: otherwise which cell index they are.
    index: int = attr.ib()

    #: Note that source contains a list of strings,
    #:  but the strings (always?) have a new line at the end if they should have one at all.
    source: Tuple[str, ...] = attr.ib(converter=optional_tuple)

    #: Not currently filed?
    output: Optional[Tuple[str, ...]] = attr.ib(converter=optional_tuple)

    execution_count: Optional[int] = attr.ib(default=None)
    # metadata: Optional[Dict[str, Any]] = attr.ib(default=None, converter=frozen_dict)

    # occurs in higher versions of nbformat
    id: Optional[str] = attr.ib(default=None)

    @property
    def joined_source(self) -> str:
        return "".join(self.source)

    @property
    def complete_source(self) -> str:
        return f"{self.cell_type}::::{self.joined_source}"

    # def __eq__(self, o):
    #     if not isinstance(o, JupyterCell):
    #         return False

    #     return self.source == o.source and self.index == o.index and self.cell_type == o.cell_type

    # def __hash__(self):
    #     return (self.cell_type, self.index, self.source)


@attr.dataclass
class NotebookContents:
    cells: List[JupyterCell]

    # TODO: Make sure cells are in the right order

    def _cell_insides(self):
        return [(x.index, x.cell_type, x.source) for x in self.cells]

    def content_equals(self, other: "NotebookContents") -> bool:
        return self._cell_insides() == other._cell_insides()


@attr.dataclass
class Movement:
    previous: int
    current: int


@attr.dataclass
class CellMovements:
    movements: List[Movement]



import os
import sys
import tempfile

from loguru import logger

from jupyter_ascending._environment import LOG_LEVEL
from jupyter_ascending._environment import SHOW_TO_STDOUT


def setup_logger():
    log_file = os.path.join(tempfile.gettempdir(), "jupyter_ascending", "log.log")
    print(f"Logging Jupyter Ascending logs to {log_file}")

    config = {
        "handlers": [{"sink": log_file, "serialize": False, "level": LOG_LEVEL}],
    }

    if SHOW_TO_STDOUT:
        config["handlers"].append({"sink": sys.stdout, "format": "{time} - {message}", "level": LOG_LEVEL})
    else:
        # Always display warning-and-up output
        config["handlers"].append({"sink": sys.stdout, "format": "{time} - {message}", "level": "WARNING"})

    logger.configure(**config)  # type: ignore



import json
import os
import socket
import urllib.request
from contextlib import closing

import ipykernel  # type: ignore
from notebook import notebookapp  # type: ignore


def get_name_from_python():
    """
    Returns the absolute path of the Notebook or None if it cannot be determined

    BUG: I think it currently doesn't _really_ return the absolute path, but it returns enough of it to be good I think.

    NOTE: works only when the security is token-based or there is also no password
    """
    connection_file = os.path.basename(ipykernel.get_connection_file())
    kernel_id = connection_file.split("-", 1)[1].split(".")[0]

    for srv in notebookapp.list_running_servers():
        try:
            if srv["token"] == "" and not srv["password"]:  # No token and no password, ahem...
                req = urllib.request.urlopen(srv["url"] + "api/sessions")
            else:
                req = urllib.request.urlopen(srv["url"] + "api/sessions?token=" + srv["token"])
            sessions = json.load(req)
            for sess in sessions:
                if sess["kernel"]["id"] == kernel_id:
                    return os.path.join(srv["notebook_dir"], sess["notebook"]["path"])
        except:
            pass  # There may be stale entries in the runtime directory

    return ""


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("localhost", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]



from typing import TypeVar

import attr
from jsonrpcclient import request
from jsonrpcclient.exceptions import ReceivedNon2xxResponseError
from requests.exceptions import ConnectionError  # type: ignore

from jupyter_ascending._environment import EXECUTE_HOST_URL
from jupyter_ascending.handlers.server_extension import perform_notebook_request
from jupyter_ascending.json_requests import JsonBaseRequest

GenericJsonRequest = TypeVar("GenericJsonRequest", bound=JsonBaseRequest)


class RequestFailure(Exception):
    pass


def request_notebook_command(json_request: GenericJsonRequest):
    """This is a command to be used by the client libraries to send a command to this server.

    It calls unpacks the JsonRequest and calls `perform_notebook_request` defined above."""
    try:
        result = request(
            EXECUTE_HOST_URL,
            perform_notebook_request.__name__,
            command_name=type(json_request).__name__,
            notebook_path=json_request.file_name,
            data=attr.asdict(json_request),
        )

        if not result.data.ok:
            raise RequestFailure(f"JSONRPC request returned as failure: {result}")

        if not result.data.result["success"]:
            raise RequestFailure(
                f"Connection to server successful, but got error message from server: {result.data.result['error']}"
            )

    except ConnectionError as e:
        raise RequestFailure(f"Unable to connect to server. Perhaps notebook is not running? {e}")
    except ReceivedNon2xxResponseError as e:
        raise RequestFailure(
            f"Unable to process request. Is jupyter-ascending installed in the server's python environment? Perhaps something else is running on this port? {e}"
        )



import time

from loguru import logger

import jupyter_ascending.handlers.server_extension
from jupyter_ascending._environment import SYNC_EXTENSION
from jupyter_ascending.handlers import jupyter_notebook
from jupyter_ascending.logger import setup_logger
from jupyter_ascending.utils import get_name_from_python


@logger.catch
def load_ipython_extension(ipython):
    """This is the specially named function that Jupyter will call to load a notebook extension."""
    set_everything_up()


def set_everything_up():
    # Note that this is also called from javascript after a kernel restart

    logger.info("Loading Ipython...")
    setup_logger()

    # Start the server if it's the right name.
    notebook_name = get_name_from_python()
    logger.info("IPYTHON: Loading {notebook}", notebook=notebook_name)

    if f".{SYNC_EXTENSION}.ipynb" not in notebook_name:
        logger.info("IPYTHON: Not loading {notebook} because name does not match", notebook=notebook_name)
        return

    logger.info("IPYTHON LOAD: " + time.ctime() + ": " + notebook_name)
    jupyter_notebook.start_notebook_server_in_thread(notebook_name)


def load_jupyter_server_extension(ipython):
    """This is the specially named function that Jupyter will call to load a server extension."""
    setup_logger()
    ipython.log.info("LOADING JUPYTER ASCENDING SERVER PLUGIN")
    logger.info("SERVER LOAD: " + time.ctime())

    jupyter_ascending.handlers.server_extension.load_extension(ipython)



"""
from typing import Optional

from attr import dataclass


@dataclass
class JsonBaseRequest:
    file_name: str


@dataclass
class ExecuteRequest(JsonBaseRequest):
    cell_index: int
    contents: Optional[str]


@dataclass
class SyncRequest(JsonBaseRequest):
    contents: str


@dataclass
class GetStatusRequest(JsonBaseRequest):
    pass


@dataclass
class FocusCellRequest(JsonBaseRequest):
    cell_index: int
"""



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) tjdevries
# Distributed under the terms of the Modified BSD License.


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "nbextension/static",
            "dest": "jupyter_ascending",
            "require": "jupyter_ascending/extension",
        }
    ]



import attr

from jupyter_ascending.notebook.data_types import JupyterCell
from jupyter_ascending.notebook.data_types import NotebookContents
from jupyter_ascending.notebook.evolve import evolve_cell_source
from jupyter_ascending.notebook.evolve import evolve_cell_type
from jupyter_ascending.notebook.evolve import evolve_notebook_cells
from jupyter_ascending.notebook.merge import OpCodes
from jupyter_ascending.notebook.merge import opcode_merge_cell_contents

SIMPLE_CONTENTS = NotebookContents(cells=[JupyterCell(cell_type="code", index=0, source=["x = 1; x"], output=["1"])])

THREE_CELL_CONTENTS = NotebookContents(
    cells=[
        JupyterCell(cell_type="code", index=0, source=["x = 1; x"], output=["1"]),
        JupyterCell(cell_type="code", index=1, source=["y = 2; y"], output=["2"]),
        JupyterCell(cell_type="code", index=2, source=["z = 3; z"], output=["3"]),
    ]
)


def _insert_notebook_cell(notebook: NotebookContents, new_cell: JupyterCell) -> NotebookContents:
    # TODO: Should this be moved somewhere?
    original_cells = notebook.cells

    new_cells = []
    for cell in original_cells:
        if cell.index < new_cell.index:
            new_cells.append(cell)
        else:
            if new_cell not in new_cells:
                new_cells.append(new_cell)

            new_cells.append(attr.evolve(cell, index=cell.index + 1))

    return evolve_notebook_cells(notebook, new_cells)


def test_passing_same_contents_means_same_output():
    opcodes = opcode_merge_cell_contents(SIMPLE_CONTENTS, SIMPLE_CONTENTS)

    assert len(opcodes) == 1
    assert opcodes[0].op_code == OpCodes.EQUAL


def test_change_one_value():
    new_contents = evolve_cell_source(SIMPLE_CONTENTS, 0, ["x = 2; x"])
    opcodes = opcode_merge_cell_contents(SIMPLE_CONTENTS, new_contents)

    assert len(opcodes) == 1
    assert opcodes[0].op_code == OpCodes.REPLACE


def test_change_one_of_multiple_values():
    new_contents = evolve_cell_source(THREE_CELL_CONTENTS, 0, ["x = 4; x"])

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)

    assert len(opcodes) == 2
    assert opcodes[0].op_code == OpCodes.REPLACE
    assert opcodes[1].op_code == OpCodes.EQUAL


def test_reorder_multiple_values():
    first_cell = attr.evolve(THREE_CELL_CONTENTS.cells[1], index=0, output=None)
    second_cell = attr.evolve(THREE_CELL_CONTENTS.cells[0], index=1, output=None)

    new_contents = NotebookContents(cells=[first_cell, second_cell, THREE_CELL_CONTENTS.cells[2]])
    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)
    assert len(opcodes) == 5

    assert [x.op_code for x in opcodes] == [
        OpCodes.INSERT,
        OpCodes.EQUAL,
        OpCodes.DELETE,
        OpCodes.EQUAL,
        OpCodes.COPY_OUTPUT,
    ]

    assert opcodes[-1].current_final_idx == 1
    assert opcodes[-1].updated_final_idx == 0


def test_remove_one_cell():
    # Notice that there isn't [2], and the output is removed
    cells_with_one_removed = [
        attr.evolve(THREE_CELL_CONTENTS.cells[0], output=None),
        attr.evolve(THREE_CELL_CONTENTS.cells[1], output=None),
    ]
    new_contents = NotebookContents(cells=cells_with_one_removed)

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)

    assert len(opcodes) == 2
    assert [x.op_code for x in opcodes] == [OpCodes.EQUAL, OpCodes.DELETE]

    assert opcodes[0].current_start_idx == 0
    assert opcodes[0].current_final_idx == 2
    assert opcodes[0].updated_start_idx == 0
    assert opcodes[0].updated_final_idx == 2


def test_remove_one_cell_beginning():
    # Notice that there isn't [0], and the output is removed
    cells_with_one_removed = [
        attr.evolve(THREE_CELL_CONTENTS.cells[1], output=None, index=0),
        attr.evolve(THREE_CELL_CONTENTS.cells[2], output=None, index=1),
    ]
    new_contents = NotebookContents(cells=cells_with_one_removed)

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)
    assert len(opcodes) == 2
    assert [x.op_code for x in opcodes] == [OpCodes.DELETE, OpCodes.EQUAL]

    assert opcodes[0].current == (0, 1)
    assert opcodes[1].current == (1, 3)


def test_remove_one_cell_and_update_another():
    # Updated source for index=1
    modified_source = ["y = 3; y"]

    cells_with_one_removed = [
        attr.evolve(THREE_CELL_CONTENTS.cells[0], output=None, index=0),
        attr.evolve(THREE_CELL_CONTENTS.cells[1], source=modified_source, output=None, index=1),
    ]
    new_contents = NotebookContents(cells=cells_with_one_removed)

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)
    assert [x.op_code for x in opcodes] == [OpCodes.EQUAL, OpCodes.REPLACE]

    assert opcodes[0].current == (0, 1)
    assert opcodes[0].updated == (0, 1)

    # Note: The replace takes the two cells and then replaces down to only one
    assert opcodes[1].current == (1, 3)
    assert opcodes[1].updated == (1, 2)


def test_insert_a_new_cell():
    inserted_cell = JupyterCell(cell_type="code", index=2, source=["print('hello world')"], output=None)
    new_contents = _insert_notebook_cell(THREE_CELL_CONTENTS, inserted_cell)

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)
    assert [x.op_code for x in opcodes] == [OpCodes.EQUAL, OpCodes.INSERT, OpCodes.EQUAL]

    assert opcodes[1].current == (2, 2)
    assert opcodes[1].updated == (2, 3)


def test_insert_a_new_cell_and_update_another():
    modified_source = ["y = 3; y"]
    inserted_cell = JupyterCell(cell_type="code", index=2, source=["print('hello world')"], output=None)

    new_contents = evolve_cell_source(_insert_notebook_cell(THREE_CELL_CONTENTS, inserted_cell), 1, modified_source)

    opcodes = opcode_merge_cell_contents(THREE_CELL_CONTENTS, new_contents)
    assert [x.op_code for x in opcodes] == [
        OpCodes.EQUAL,
        OpCodes.REPLACE,
        OpCodes.EQUAL,
    ]

    assert opcodes[1].current == (1, 2)
    assert opcodes[1].updated == (1, 3)


def test_changing_cell_type_sends_replace():
    new_contents = evolve_cell_type(SIMPLE_CONTENTS, 0, "markdown")
    opcodes = opcode_merge_cell_contents(SIMPLE_CONTENTS, new_contents)

    assert new_contents != SIMPLE_CONTENTS
    assert len(opcodes) == 1
    assert opcodes[0].op_code == OpCodes.REPLACE



import abc
from collections import defaultdict
from difflib import SequenceMatcher
from difflib import get_close_matches
from enum import Enum
from typing import Dict
from typing import List
from typing import Set
from typing import Tuple
from typing import Union

import attr
import edlib

from jupyter_ascending.notebook.data_types import CellMovements
from jupyter_ascending.notebook.data_types import JupyterCell
from jupyter_ascending.notebook.data_types import Movement
from jupyter_ascending.notebook.data_types import NotebookContents
from jupyter_ascending.notebook.evolve import evolve_notebook_cells

Number = Union[int, float]


def _get_raw_contents(notebook: NotebookContents) -> Tuple[str, ...]:
    return tuple(x.complete_source for x in notebook.cells)


class OpCodes(Enum):
    EQUAL = "equal"
    INSERT = "insert"
    DELETE = "delete"
    REPLACE = "replace"

    # Custom actions for us.
    COPY_OUTPUT = "copy_output"


@attr.dataclass
class OpCodeAction:
    op_code: OpCodes
    current_start_idx: int
    current_final_idx: int
    updated_start_idx: int
    updated_final_idx: int

    @property
    def current(self) -> Tuple[int, int]:
        return (self.current_start_idx, self.current_final_idx)

    @property
    def updated(self) -> Tuple[int, int]:
        return (self.updated_start_idx, self.updated_final_idx)


def opcode_merge_cell_contents(
    current_notebook: NotebookContents, updated_notebook: NotebookContents
) -> List[OpCodeAction]:
    raw_current_contents = _get_raw_contents(current_notebook)
    raw_updated_contents = _get_raw_contents(updated_notebook)

    sequence_matcher = SequenceMatcher(None, raw_current_contents, raw_updated_contents)

    current_cells_consumed: Set[int] = set()
    cells_to_update_text: List[int] = []

    opcode_results = []
    for (
        opcode,
        current_start_idx,
        current_final_idx,
        updated_start_idx,
        updated_final_idx,
    ) in sequence_matcher.get_opcodes():

        if opcode in {"replace", "insert"}:
            cells_to_update_text.extend(range(updated_start_idx, updated_final_idx))

        if opcode not in {"delete"}:
            current_cells_consumed.update(range(current_start_idx, current_final_idx))

        opcode_results.append(
            OpCodeAction(OpCodes(opcode), current_start_idx, current_final_idx, updated_start_idx, updated_final_idx)
        )

    current_cells_remaining = set(range(len(raw_current_contents))) - current_cells_consumed

    for updated_cell_index in cells_to_update_text:
        closest_match = get_close_matches(
            raw_updated_contents[updated_cell_index], [raw_current_contents[x] for x in current_cells_remaining], n=1
        )

        if closest_match:
            match = closest_match[0]

            for current_cell_idx in current_cells_remaining:
                if raw_current_contents[current_cell_idx] == match:
                    current_cells_remaining.remove(current_cell_idx)
                    # TODO: Maybe it makes sense to do plus 1 here. Off-by-one is hard
                    opcode_results.append(
                        OpCodeAction(
                            OpCodes("copy_output"),
                            current_cell_idx,
                            current_cell_idx,
                            updated_cell_index,
                            updated_cell_index,
                        )
                    )
                    break

    return opcode_results


def merge_cell_contents(
    current_notebook: NotebookContents, updated_notebook: NotebookContents
) -> Tuple[NotebookContents, CellMovements]:
    # TODO: change name of current_notebook to be remote notebook? some other name that makes it obvious
    if current_notebook.content_equals(updated_notebook):
        return current_notebook, CellMovements(movements=[])

    current_cell_stack = [x for x in current_notebook.cells]
    updated_cell_stack = [x for x in updated_notebook.cells]

    movements = []
    final_cells = []

    # 1. Check if we have cells that are exactly the same source
    #   We just update to the new index
    current_cells_to_remove = []
    for current_cell in current_cell_stack:
        for updated_cell in updated_cell_stack:
            if current_cell.source == updated_cell.source:

                # Create movements
                if current_cell.index != updated_cell.index:
                    movements.append(Movement(previous=current_cell.index, current=updated_cell.index))

                final_cells.append(attr.evolve(current_cell, index=updated_cell.index))

                # Remove at the end, since it's not good to modify iterators while iterating
                current_cells_to_remove.append(current_cell)

                updated_cell_stack.remove(updated_cell)
                break

    for cell in current_cells_to_remove:
        current_cell_stack.remove(cell)

    # 2. gather up all the differences between the remaining cells
    distancer = LevenshteinDistance

    distance_between_cells: Dict[JupyterCell, List[CellDistance]] = defaultdict(list)
    for current_cell in current_cell_stack:
        # TODO: What if all the distances are bad?
        # TODO: maybe use a different measure? like some confidence that they're the same
        for updated_cell in updated_cell_stack:
            if current_cell.source == updated_cell.source:
                assert False, f"{current_cell} / {updated_cell}::\n{final_cells}"

            distance = distancer.find_distance(current_cell.joined_source, updated_cell.joined_source)
            distance_between_cells[current_cell].append(CellDistance(distance, updated_cell))

        distance_between_cells[current_cell] = list(
            sorted(distance_between_cells[current_cell], key=distancer.sort_function)
        )

    # 3. Find most closely related cells (so, small distance)
    #       If I run out of updated_cell_stack, then we're done. No more cells to give
    #       If I run out of current_cell_stack, then we need to delete those (however that looks).

    def find_most_likely_cell(x):
        distance = distance_between_cells[x]

        return distancer.sort_function(distance[0])

    # TODO: This might be way too many loops?
    while current_cell_stack and distance_between_cells and updated_cell_stack:
        sorted_cell_list = sorted(current_cell_stack, key=find_most_likely_cell)
        current_cell = sorted_cell_list[0]

        cell_distances = distance_between_cells.pop(current_cell)
        best_updated_cell = next(x.cell for x in cell_distances if x.cell in updated_cell_stack)

        assert current_cell in current_cell_stack
        current_cell_stack.remove(current_cell)

        assert best_updated_cell in updated_cell_stack
        updated_cell_stack.remove(best_updated_cell)

        evolved_cell = attr.evolve(current_cell, index=best_updated_cell.index, source=best_updated_cell.source)
        final_cells.append(evolved_cell)

    # Any other updated cells we have must have been inserted.
    #   We can simply insert them into the final cells
    #   (Unless there is something I missed here...)
    while updated_cell_stack:
        final_cells.append(updated_cell_stack.pop())

    # Should have no more updated cells to input
    assert not updated_cell_stack

    return evolve_notebook_cells(current_notebook, final_cells), CellMovements(movements=movements)


@attr.dataclass
class CellDistance:
    distance: Number
    cell: JupyterCell


class BaseStringDistancer(abc.ABC):
    @staticmethod
    def find_distance(string_1: str, string_2: str) -> Number:
        raise NotImplementedError

    @staticmethod
    def sort_function(distance: CellDistance) -> Number:
        raise NotImplementedError


class LevenshteinDistance(BaseStringDistancer):
    @staticmethod
    def find_distance(string_1: str, string_2: str) -> int:
        # TODO: Maybe use jaro winkler instead
        return edlib.align(string_1, string_2)["editDistance"]

    @staticmethod
    def sort_function(distance: CellDistance) -> Number:
        return distance.distance



import argparse
from pathlib import Path

from loguru import logger

from jupyter_ascending.json_requests import RestartRequest
from jupyter_ascending.logger import setup_logger
from jupyter_ascending.requests.client_lib import request_notebook_command


def send(file_name: str):
    logger.info(f"Restarting kernel in file: {file_name}...")
    file_name = str(Path(file_name).absolute())

    request_obj = RestartRequest(file_name=file_name)
    request_notebook_command(request_obj)

    logger.info("... Complete")


if __name__ == "__main__":
    setup_logger()
    parser = argparse.ArgumentParser()

    parser.add_argument("--filename", help="Filename to send")

    arguments = parser.parse_args()
    send(arguments.filename)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jupyter_ascending_patched',
 'jupyter_ascending_patched.handlers',
 'jupyter_ascending_patched.nbextension',
 'jupyter_ascending_patched.notebook',
 'jupyter_ascending_patched.requests',
 'jupyter_ascending_patched.scripts',
 'jupyter_ascending_patched.tests']

package_data = \
{'': ['*'],
 'jupyter_ascending_patched': ['labextension/*'],
 'jupyter_ascending_patched.nbextension': ['static/*']}

install_requires = \
['attrs>=19.0.0',
 'edlib>=1.3.8,<2.0.0',
 'ipywidgets>=7.5.1,<8.0.0',
 'jsonrpcclient[requests]>=3.3.5,<4.0.0',
 'jsonrpcserver[requests]>=4.1.2,<5.0.0',
 'jupytext>=1.3.4,<2.0.0',
 'loguru>=0.4.1',
 'toml>=0.10.0']

setup_kwargs = {
    'name': 'jupyter-ascending-patched',
    'version': '0.1.23.1',
    'description': '',
    'long_description': None,
    'author': 'Josh Albrecht',
    'author_email': 'joshalbrecht@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from functools import wraps
from http.server import BaseHTTPRequestHandler
from typing import Callable

from jsonrpcserver import dispatch
from jsonrpcserver import methods
from loguru import logger


def _wrap_request(f: Callable, start_msg: str, close_msg: str):
    @wraps(f)
    @logger.catch
    def wrapper(*args, **kwargs):
        logger.debug("{}: {}", start_msg, f.__name__)

        result = f(*args, **kwargs)

        logger.debug("{}: {}", close_msg, result)

        return result

    return wrapper


class ServerMethods:
    """
    Wrapper to make some things a bit nicer around jsonrpcserver.methods.Methods

    Basically our own version of jsonrpcserver.methods.Methods, wrapping each method with auto
    logging and error catching so that you don't have to remember to do that.
    """

    def __init__(self, start_msg: str, close_msg: str):
        self.items = {}
        self.start_msg = start_msg
        self.close_msg = close_msg

    def add(self, f: Callable) -> Callable:
        self.items[f.__name__] = f

        return _wrap_request(f, self.start_msg, self.close_msg)


def generate_request_handler(name: str, methods: ServerMethods) -> BaseHTTPRequestHandler:
    """Build a handler to respond to HTTP POST requests containing JSON-RPC messages.

    Will call jsonrpcserver.dispatch to dispatch the request to the appropriate handler.

    TODO: why this weird construction vs a simple subclass?
        - to be able to specify a custom class name, i think. but why do we need that?
    """

    @logger.catch
    def do_POST(self):
        # Process request
        request = self.rfile.read(int(self.headers["Content-Length"])).decode()
        logger.info("{} processing request:\n\t\t{}", name, request)

        # Dispatch the RPC request to the right function and get the function's response.
        response = dispatch(request, methods=methods)

        logger.info("Got Response:\n\t\t{}", response)

        # Return response
        self.send_response(response.http_status)
        self.send_header("Content-Type", "application/json")
        self.end_headers()
        self.wfile.write(str(response).encode())

    def log_message(self, format, *args):
        logger.debug(args)

    return type(
        f"{name}RequestHandler",
        (BaseHTTPRequestHandler,),
        {"allow_reuse_address": True, "do_POST": do_POST, "log_message": log_message},
    )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) tjdevries.
# Distributed under the terms of the Modified BSD License.

# NOTE: Must keep in sync w/ pyproject.toml to publish
version_info = (0, 1, 23, 1)
__version__ = ".".join(map(str, version_info))
